import { Controller, Request, Post, UseGuards, Body, Res, Delete, Patch, UnauthorizedException } from '@nestjs/common';
import { AuthService } from './auth.service';
import { UsersService } from 'src/users/users.service';
import { RegisterUserDto } from './dto/register-user.dto';
import { Response } from 'express';
import { JwtRefreshGuard } from './guards/jwt-refresh.guard';
import { OfflineGuard } from './guards/offline.guard';
import { UserEntity } from 'src/users/entities/users.entity';
import { QueryFailedError } from 'typeorm';
import { PinoLogger } from 'nestjs-pino';
import { LoginDto } from './dto/login.dto';
import { AuthenticationDto } from './dto/authentication.dto';
import { AppConfig } from 'src/asset/config/app-config';
import { JwtMixedGuard } from './guards/jwt-mixed.guard';

@Controller('auth')
export class AuthController {
  constructor(
    private auth: AuthService,
    private users: UsersService,
    private logger: PinoLogger,
  ) { }

  @UseGuards(OfflineGuard)
  @Post('login')
  async login(
    @Request() req,
    @Res({ passthrough: true }) response: Response,
    @Body() body: LoginDto,
  ) {
    let data: AuthenticationDto | null;
    try {
      data = await this.auth.login(body);
      if (!data.access_token || body.remember_me && (!data.refresh_token || !data.refresh_exp)) {
        response.statusCode = 500;
        return {
          success: false,
          error_message: 'Something went wrong with tokens while logging in.',
        };
      }
    } catch (err) {
      this.logger.error({
        class: AuthController.name,
        method: this.login.name,
        user: req.user,
        msg: 'Failed to login.',
        error: err,
      });

      if (err instanceof UnauthorizedException) {
        response.statusCode = 401;
        return {
          success: false,
          error_message: 'Invalid credentials.',
        };
      }

      response.statusCode = 500;
      return {
        success: false,
        error_message: 'Something went wrong while logging in.',
      };
    }

    response.cookie('Authentication', data.access_token, {
      httpOnly: true,
      secure: true,
      expires: new Date(data.exp),
      sameSite: 'strict',
    });

    if (body.remember_me) {
      response.cookie('Refresh', data.refresh_token, {
        httpOnly: true,
        secure: true,
        expires: new Date(data.refresh_exp),
        sameSite: 'strict',
      });
    }

    this.logger.info({
      class: AuthController.name,
      method: this.login.name,
      user: req.user,
      access_token: data.access_token,
      refresh_token: data.refresh_token,
      remember_me: body.remember_me,
      msg: 'User logged in.',
    });

    return {
      success: true,
    };
  }

  @UseGuards(JwtMixedGuard)
  @Delete('login')
  async logout(
    @Request() req,
    @Res({ passthrough: true }) response: Response,
  ) {
    const accessToken = req.cookies?.Authentication;
    const refreshToken = req.cookies?.Refresh;

    response.clearCookie('Authentication');
    response.clearCookie('Refresh');

    if (!accessToken && !refreshToken && !await this.auth.revoke(req.user.userId, refreshToken)) {
      // User has already logged off.
      this.logger.info({
        class: AuthController.name,
        method: this.logout.name,
        user: req.user,
        msg: 'User has already logged off based on ' + (!refreshToken ? 'cookies' : 'database'),
      });

      response.statusCode = 400;
      return {
        success: false,
        error_message: 'User has already logged off.',
      };
    }

    this.logger.info({
      class: AuthController.name,
      method: this.logout.name,
      user: req.user,
      msg: 'User logged off',
    });

    return {
      success: true,
    };
  }

  @UseGuards(JwtRefreshGuard)
  @Patch('login')
  async refresh(
    @Request() req,
    @Res({ passthrough: true }) response: Response,
  ) {
    this.logger.info({
      class: AuthController.name,
      method: this.refresh.name,
      user: req.user,
      refresh_token: req.cookies.Refresh,
      msg: 'Attempting to renew access token.',
    });

    const results = await this.auth.verify(req.cookies.Authentication, req.cookies.Refresh);

    if (results.validation === false) {
      this.logger.info({
        class: AuthController.name,
        method: this.refresh.name,
        user: req.user,
        refresh_token: req.cookies.Refresh,
        msg: 'Refresh token is invalid. Access token is not refreshing.',
      });

      response.statusCode = 400;
      return {
        success: false,
        error_message: 'Refresh token is invalid.',
      };
    }

    const data = await this.auth.renew(req.user);

    response.cookie('Authentication', data.access_token, {
      httpOnly: true,
      secure: true,
      expires: new Date(data.exp),
      sameSite: 'strict',
    });
    this.logger.debug({
      class: AuthController.name,
      method: this.refresh.name,
      user: req.user,
      access_token: data.access_token,
      msg: 'Updated Authentication cookie for access token.',
    });

    return { success: true };
  }

  @UseGuards(OfflineGuard)
  @Post('register')
  async register(
    @Request() req,
    @Res({ passthrough: true }) response: Response,
    @Body() body: RegisterUserDto,
  ) {
    if (!AppConfig.features.registration) {
      response.statusCode = 404;
      return {
        success: false,
        error_message: 'Registration disabled.',
      };
    }

    let user: UserEntity | null;
    let data: AuthenticationDto | null;
    try {
      const { user_login, user_name, password } = body;
      user = await this.users.register(user_login.toLowerCase(), user_name, password, true);
      this.logger.info({
        class: AuthController.name,
        method: this.register.name,
        user: req.user,
        msg: 'User registered',
      });
    } catch (err) {
      if (err instanceof QueryFailedError) {
        if (err.message.includes('duplicate key value violates unique constraint "users_user_login_key"')) {
          this.logger.warn({
            class: AuthController.name,
            method: this.register.name,
            user: req.user,
            msg: 'Failed to register due to duplicate userLogin.',
          });

          response.statusCode = 409;
          return {
            success: false,
            error_message: 'Username already exist.',
          };
        }
      }
      this.logger.error({
        class: AuthController.name,
        method: this.register.name,
        user: req.user,
        msg: 'Failed to register.',
        error: err,
      });

      response.statusCode = 500;
      return {
        success: false,
        error_message: 'Something went wrong when creating user.',
      };
    }

    try {
      data = await this.auth.login({
        user_login: body.user_login,
        password: body.password,
        remember_me: false,
      });
      if (!data.access_token) {
        this.logger.error({
          class: AuthController.name,
          method: this.register.name,
          user: req.user,
          access_token: data.access_token,
          refresh_token: data.refresh_token,
          msg: 'Failed to generate tokens after registering.',
        });

        response.statusCode = 500;
        return {
          success: false,
          error_message: 'Something went wrong with tokens while logging in.',
        };
      }
    } catch (err) {
      this.logger.error({
        class: AuthController.name,
        method: this.register.name,
        user: req.user,
        msg: 'Failed to login after registering.',
        error: err,
      });

      // This should never happen...
      if (err instanceof UnauthorizedException) {
        response.statusCode = 401;
        return {
          success: false,
          error_message: 'Invalid credentials.',
        };
      }

      response.statusCode = 500;
      return {
        success: false,
        error_message: 'Something went wrong while logging in.',
      };
    }

    response.cookie('Authentication', data.access_token, {
      httpOnly: true,
      secure: true,
      expires: new Date(data.exp),
      sameSite: 'strict',
    });

    return {
      success: true,
    };
  }

  @Post('validate')
  async validate(
    @Request() req,
    @Res({ passthrough: true }) response: Response,
  ) {
    try {
      const accessToken = req.cookies['Authentication'];
      const refreshToken = req.cookies['Refresh'];
      const verification = await this.auth.verify(accessToken, refreshToken);

      return {
        success: true,
        ...verification,
      };
    } catch (err) {
      response.statusCode = 500;
      return {
        success: false,
        error_message: err,
      };
    }
  }
}