import * as crypto from 'crypto';
import * as moment from "moment";
import { Injectable } from '@nestjs/common';
import { ConfigService } from '@nestjs/config';
import { JwtService } from '@nestjs/jwt';
import { InjectRepository } from '@nestjs/typeorm';
import { UUID } from 'crypto';
import { UserEntity } from 'src/users/entities/users.entity';
import { Repository } from 'typeorm';
import { AuthRefreshTokenEntity } from './entities/auth.refresh-token.entity';
import { PinoLogger } from 'nestjs-pino';

@Injectable()
export class AuthRefreshService {
  constructor(
    private jwts: JwtService,
    private config: ConfigService,
    @InjectRepository(AuthRefreshTokenEntity)
    private authRefreshTokenRepository: Repository<AuthRefreshTokenEntity>,
    private logger: PinoLogger,
  ) { }

  async generate(user: UserEntity) {
    const now = new Date();
    const expirationTime = parseInt(this.config.getOrThrow<string>('AUTH_JWT_REFRESH_TOKEN_EXPIRATION_MS'));
    const expiration = moment(now).add(expirationTime, 'ms').toDate();
    const refreshToken = await this.jwts.signAsync(
      {
        username: user.userLogin,
        sub: user.userId,
        iat: Math.floor(now.getTime() / 1000),
        nbf: Math.floor(now.getTime() / 1000) - 5 * 60,
        exp: Math.floor(expiration.getTime() / 1000),
      },
      {
        secret: this.config.getOrThrow<string>('AUTH_JWT_REFRESH_TOKEN_SECRET'),
      }
    );

    this.logger.debug({
      class: AuthRefreshService.name,
      method: this.generate.name,
      user,
      refresh_token: refreshToken,
      exp: expiration,
      msg: 'Generated a new refresh token.',
    });

    this.authRefreshTokenRepository.insert({
      tokenHash: this.hash(refreshToken),
      userId: user.userId,
      exp: expiration
    });

    this.logger.debug({
      class: AuthRefreshService.name,
      method: this.generate.name,
      user,
      refresh_token: refreshToken,
      exp: expiration,
      msg: 'Inserted the new refresh token into the database.',
    });

    return {
      refresh_token: refreshToken,
      exp: expiration.getTime(),
    }
  }

  async get(
    refreshToken: string,
    userId: UUID,
  ): Promise<{ tokenHash: string, userId: UUID, exp: Date }> {
    if (!refreshToken || !userId) {
      return null;
    }

    return await this.authRefreshTokenRepository.findOneBy({
      tokenHash: this.hash(refreshToken),
      userId: userId,
    });
  }

  private hash(refreshToken: string): string {
    const buffer = Buffer.from(refreshToken, 'utf8');
    return crypto.createHash('sha256').update(buffer).digest('base64');
  }

  async revoke(userId: UUID, refreshToken: string) {
    if (!userId || !refreshToken) {
      return null;
    }

    return await this.authRefreshTokenRepository.delete({
      userId,
      tokenHash: this.hash(refreshToken),
    });
  }

  async validate(
    refreshToken: string,
    userId: UUID,
  ): Promise<boolean> {
    const refresh = await this.get(refreshToken, userId);
    return refresh && refresh.exp.getTime() > new Date().getTime();
  }

  async verify(
    refreshToken: string
  ): Promise<any> {
    return await this.jwts.verifyAsync(refreshToken,
      {
        secret: this.config.getOrThrow<string>('AUTH_JWT_REFRESH_TOKEN_SECRET'),
      }
    );
  }
}
