import { InjectQueue } from '@nestjs/bullmq';
import { Body, Controller, Delete, Get, Patch, Post, Put, Request, Res, UseGuards } from '@nestjs/common';
import { Response } from 'express';
import { Queue } from 'bullmq';
import { PinoLogger } from 'nestjs-pino';
import { BooksService } from 'src/books/books.service';
import { BookSearchResultDto } from 'src/providers/dto/book-search-result.dto';
import { SeriesService } from 'src/series/series.service';
import { QueryFailedError } from 'typeorm';
import { UpdateBookDto } from 'src/books/dto/update-book.dto';
import { UpdateBookOriginDto } from 'src/books/dto/update-book-origin.dto';
import { LibraryService } from './library.service';
import { JwtAccessGuard } from 'src/auth/guards/jwt-access.guard';
import { SeriesDto } from 'src/series/dto/series.dto';
import { DeleteBookStatusDto } from 'src/books/dto/delete-book-status.dto';
import { CreateBookStatusDto } from 'src/books/dto/create-book-status.dto';
import { JwtAccessAdminGuard } from 'src/auth/guards/jwt-access.admin.guard';
import { BookOriginDto } from 'src/books/dto/book-origin.dto';
import { CreateSeriesDto } from 'src/series/dto/create-series.dto';

@UseGuards(JwtAccessGuard)
@Controller('library')
export class LibraryController {
  constructor(
    private readonly books: BooksService,
    private readonly series: SeriesService,
    private readonly library: LibraryService,
    @InjectQueue('library') private readonly jobs: Queue,
    private readonly logger: PinoLogger,
  ) { }

  @Get('series')
  async getSeries(
    @Request() req,
  ) {
    return {
      success: true,
      data: await this.series.getAllSeries(),
    };
  }

  @Post('series')
  async createSeries(
    @Request() req,
    @Body() body: CreateSeriesDto,
    @Res({ passthrough: true }) response: Response,
  ) {
    try {
      await this.library.addSeries(body);

      return {
        success: true,
      };
    } catch (err) {
      if (err instanceof QueryFailedError) {
        if (err.driverError.code == '23505') {
          this.logger.warn({
            class: LibraryController.name,
            method: this.createSeries.name,
            user: req.user,
            body: body,
            msg: 'Failed to create a series.',
            error: err,
          });

          response.statusCode = 409;
          return {
            success: false,
            error_message: 'Series already exists.',
          };
        }
      }

      this.logger.error({
        class: LibraryController.name,
        method: this.createSeries.name,
        user: req.user,
        body: body,
        msg: 'Failed to create a series.',
        error: err,
      });

      response.statusCode = 500;
      return {
        success: false,
        error_message: 'Something went wrong.',
      };
    }
  }

  @Patch('series')
  async updateSeries(
    @Request() req,
    @Body() body: CreateSeriesDto,
    @Res({ passthrough: true }) response: Response,
  ) {
    try {
      await this.library.updateSeries(body);

      return {
        success: true,
      };
    } catch (err) {
      if (err instanceof QueryFailedError) {
        if (err.driverError.code == '23505') {
          this.logger.warn({
            class: LibraryController.name,
            method: this.updateSeries.name,
            user: req.user,
            msg: 'Failed to update a series.',
            error: err,
          });

          // Subscription already exist.
          response.statusCode = 409;
          return {
            success: false,
            error_message: 'Series subscription already exists.',
          };
        }
      }

      this.logger.error({
        class: LibraryController.name,
        method: this.updateSeries.name,
        user: req.user,
        body: body,
        msg: 'Failed to update a series.',
        error: err,
      });

      response.statusCode = 500;
      return {
        success: false,
        error_message: 'Something went wrong.',
      };
    }
  }

  @UseGuards(JwtAccessAdminGuard)
  @Delete('series')
  async deleteSeries(
    @Request() req,
    @Body() body: SeriesDto,
    @Res({ passthrough: true }) response: Response,
  ) {
    try {
      const del = await this.series.deleteSeries(body);
      return {
        success: del && del.affected > 0,
      };
    } catch (err) {
      if (err instanceof QueryFailedError) {
        if (err.driverError.code == '23503') {
          this.logger.warn({
            class: LibraryController.name,
            method: this.deleteSeries.name,
            user: req.user,
            body: body,
            msg: 'Failed to delete a series.',
            error: err,
          });

          response.statusCode = 404;
          return {
            success: false,
            error_message: 'The series does not exist.',
          };
        }
      }

      this.logger.error({
        class: LibraryController.name,
        method: this.deleteSeries.name,
        user: req.user,
        body: body,
        msg: 'Failed to delete a series.',
        error: err,
      });

      response.statusCode = 500;
      return {
        success: false,
        error_message: 'Something went wrong.',
      };
    }
  }

  @Get('series/subscriptions')
  async getSeriesSubscriptions(
    @Request() req,
  ) {
    return {
      success: true,
      data: await this.series.getSeriesSubscribedBy(req.user.userId),
    };
  }

  @Post('series/subscribe')
  async subscribe(
    @Request() req,
    @Body() body: SeriesDto,
    @Res({ passthrough: true }) response: Response,
  ) {
    try {
      await this.library.addSubscription({
        ...body,
        userId: req.user.userId,
      });

      return {
        success: true,
      };
    } catch (err) {
      if (err instanceof QueryFailedError) {
        if (err.driverError.code == '23505') {
          this.logger.warn({
            class: LibraryController.name,
            method: this.subscribe.name,
            user: req.user,
            body: body,
            msg: 'Failed to subscribe to a series.',
            error: err,
          });

          // Subscription already exists.
          response.statusCode = 409;
          return {
            success: false,
            error_message: 'Series subscription already exists.',
          };
        } else if (err.driverError.code == '23503') {
          this.logger.warn({
            class: LibraryController.name,
            method: this.subscribe.name,
            user: req.user,
            body: body,
            msg: 'Failed to subscribe to a series.',
            error: err,
          });

          // Series does not exist.
          response.statusCode = 404;
          return {
            success: false,
            error_message: 'Series does not exist.',
          };
        }
      }

      this.logger.error({
        class: LibraryController.name,
        method: this.subscribe.name,
        user: req.user,
        body: body,
        msg: 'Failed to subscribe to a series.',
        error: err,
      });

      response.statusCode = 500;
      return {
        success: false,
        error_message: 'Something went wrong.',
      };
    }
  }

  @Delete('series/subscribe')
  async deleteSeriesSubscription(
    @Request() req,
    @Body() body: SeriesDto,
  ) {
    const del = await this.series.deleteSeriesSubscription({
      ...body,
      userId: req.user.userId,
    });
    return {
      success: del && del.affected > 0,
    };
  }

  @Get('books')
  async getBooksFromUser(
    @Request() req,
  ) {
    return {
      success: true,
      data: await this.library.findBooks(),
    };
  }

  @Post('books')
  async createBook(
    @Request() req,
    @Body() body: BookSearchResultDto,
    @Res({ passthrough: true }) response: Response,
  ) {
    if (body.provider && body.providerSeriesId) {
      this.logger.warn({
        class: LibraryController.name,
        method: this.createBook.name,
        user: req.user,
        body: body,
        msg: 'Failed to create book due to book being part of a series.',
      });

      response.statusCode = 400;
      return {
        success: false,
        error_message: 'This book is part of a seris. Use the series route to create a series.',
      }
    }

    try {
      return {
        success: true,
        data: await this.books.createBook(body),
      };
    } catch (err) {
      if (err instanceof QueryFailedError) {
        if (err.driverError.code == '23505') {
          this.logger.warn({
            class: LibraryController.name,
            method: this.createBook.name,
            user: req.user,
            body: body,
            msg: 'Failed to create book.',
            error: err,
          });

          // Book exists already.
          response.statusCode = 409;
          return {
            success: false,
            error_message: 'The book has already been added previously.',
          };
        } else if (err.driverError.code == '23503') {
          this.logger.warn({
            class: LibraryController.name,
            method: this.createBook.name,
            user: req.user,
            body: body,
            msg: 'Failed to create book.',
            error: err,
          });

          // Series is missing.
          response.statusCode = 404;
          return {
            success: false,
            error_message: 'Series does not exist.',
          };
        }
      }

      this.logger.error({
        class: LibraryController.name,
        method: this.createBook.name,
        user: req.user,
        body: body,
        msg: 'Failed to create a book.',
        error: err,
      });

      response.statusCode = 500;
      return {
        success: false,
        error_message: 'Something went wrong while adding the book.',
      };
    }
  }

  @Put('books')
  async updateBook(
    @Body() body: UpdateBookDto,
  ) {
    const data = { ...body };
    delete data['bookId'];

    const result = await this.books.updateBook(body.bookId, data);
    return {
      success: result && result.affected > 0,
    };
  }

  @Delete('books/origins')
  async deleteBookOrigin(
    @Body() body: BookOriginDto[],
  ) {
    const result = await this.books.deleteBookOrigin(body);
    return {
      success: result && result.affected > 0,
    };
  }

  @Put('books/origins')
  async updateBookOrigin(
    @Body() body: UpdateBookOriginDto,
  ) {
    const data = { ...body };
    delete data['bookOriginId'];

    const result = await this.books.updateBookOrigin(body.bookOriginId, data);
    return {
      success: result && result.affected > 0,
    };
  }

  @Get('books/status')
  async getBookStatus(
    @Request() req,
    @Body() body: SeriesDto,
  ) {
    return {
      success: true,
      data: await this.books.findActualBookStatusesTrackedBy(req.user.userId, body),
    };
  }

  @Put('books/status')
  async updateBookStatus(
    @Request() req,
    @Body() body: CreateBookStatusDto,
    @Res({ passthrough: true }) response: Response,
  ) {
    try {
      await this.books.updateBookStatus(body);
      return {
        success: true,
      };
    } catch (err) {
      if (err instanceof QueryFailedError) {
        if (err.driverError.code == '23503') {
          this.logger.warn({
            class: LibraryController.name,
            method: this.updateBookStatus.name,
            user: req.user,
            body: body,
            msg: 'Failed to update the user\'s status of a book.',
            error: err,
          });

          response.statusCode = 404;
          return {
            success: false,
            error_message: 'The book does not exist.',
          };
        }
      }

      this.logger.error({
        class: LibraryController.name,
        method: this.updateBookStatus.name,
        user: req.user,
        body: body,
        msg: 'Failed to update the user\'s status of a book.',
        error: err,
      });

      response.statusCode = 500;
      return {
        success: false,
        error_message: 'Something went wrong.',
      };
    }
  }

  @Delete('books/status')
  async deleteBookStatus(
    @Body() body: DeleteBookStatusDto,
  ) {
    const result = await this.books.deleteBookStatus(body);
    return {
      success: result && result.affected > 0,
    };
  }
}
